@@ -7,14 +7,21 @@ coURLan: Clean, filter, normalize, and sample URLs
 __title__ = "courlan"
 __author__ = "Adrien Barbaresi"
 __license__ = "GNU GPL v3+"
-__copyright__ = "Copyright 2020-2022, Adrien Barbaresi"
-__version__ = "0.8.3"
+__copyright__ = "Copyright 2020-2023, Adrien Barbaresi"
+__version__ = "0.9.5"
 
 
 # imports
 from .clean import clean_url, normalize_url, scrub_url
-from .core import check_url, extract_links, sample_urls
-from .filters import is_navigation_page, is_not_crawlable, lang_filter, validate_url
+from .core import check_url, extract_links
+from .filters import (
+    is_navigation_page,
+    is_not_crawlable,
+    is_valid_url,
+    lang_filter,
+    validate_url,
+)
+from .sampling import sample_urls
 from .urlstore import UrlStore
 from .urlutils import (
     extract_domain,
