@@ -9,10 +9,11 @@ import logging
 import re
 
 from typing import Optional, Union
-from urllib.parse import parse_qs, urlencode, urlparse, ParseResult
+from urllib.parse import parse_qs, quote, urlencode, urlunsplit, SplitResult
 
-from .filters import validate_url
+from .filters import is_valid_url
 from .settings import ALLOWED_PARAMS, CONTROL_PARAMS, TARGET_LANG_DE, TARGET_LANG_EN
+from .urlutils import _parse
 
 
 LOGGER = logging.getLogger(__name__)
@@ -22,6 +23,7 @@ PROTOCOLS = re.compile(r"https?://")
 SELECTION = re.compile(
     r'(https?://[^">&? ]+?)(?:https?://)|(?:https?://[^/]+?/[^/]+?[&?]u(rl)?=)(https?://[^"> ]+)'
 )
+
 MIDDLE_URL = re.compile(r"https?://.+?(https?://.+?)(?:https?://|$)")
 NETLOC_RE = re.compile(r"(?<=\w):(?:80|443)")
 
@@ -30,13 +32,23 @@ PATH1 = re.compile(r"/+")
 PATH2 = re.compile(r"^(?:/\.\.(?![^/]))+")
 
 # scrub
-LINK_TAG = re.compile(r"</?a>")
+REMAINING_MARKUP = re.compile(r"</?[a-z]{,4}?>|{.+?}")
 TRAILING_AMP = re.compile(r"/\&$")
 TRAILING_PARTS = re.compile(r'(.*?)[<>"\'\s]')
 
+# https://github.com/AdguardTeam/AdguardFilters/blob/master/TrackParamFilter/sections/general_url.txt
+# https://gitlab.com/ClearURLs/rules/-/blob/master/data.min.json
+# https://firefox.settings.services.mozilla.com/v1/buckets/main/collections/query-stripping/records
+TRACKERS_RE = re.compile(
+    r"^(?:dc|fbc|gc|twc|yc|ysc)lid|"
+    r"^(?:click|gbra|msclk|igsh|partner|wbra)id|"
+    r"^(?:ads?|mc|ga|gs|itm|mc|mkt|ml|mtm|oly|pk|utm|vero)_|"
+    r"(?:\b|_)(?:aff|affi|affiliate|campaign|cl?id|eid|ga|gl|kwd|keyword|medium|ref|referr?er|session|source|uid|xtor)"
+)
+
 
 def clean_url(url: str, language: Optional[str] = None) -> Optional[str]:
-    """Helper function: chained scrubbing and normalization"""
+    "Helper function: chained scrubbing and normalization"
     try:
         return normalize_url(scrub_url(url), False, language)
     except (AttributeError, ValueError):
@@ -44,7 +56,7 @@ def clean_url(url: str, language: Optional[str] = None) -> Optional[str]:
 
 
 def scrub_url(url: str) -> str:
-    """Strip unnecessary parts and make sure only one URL is considered"""
+    "Strip unnecessary parts and make sure only one URL is considered"
     # trim
     # https://github.com/cocrawler/cocrawler/blob/main/cocrawler/urls.py
     # remove leading and trailing white space and unescaped control chars
@@ -59,7 +71,7 @@ def scrub_url(url: str) -> str:
         url = url.replace("<![CDATA[", "")  # url = re.sub(r'^<!\[CDATA\[', '', url)
         url = url.replace("]]>", "")  # url = re.sub(r'\]\]>$', '', url)
     # markup rests
-    url = LINK_TAG.sub("", url)
+    url = REMAINING_MARKUP.sub("", url)
     # & and &amp;
     if "&amp;" in url:
         url = url.replace("&amp;", "&")
@@ -71,20 +83,20 @@ def scrub_url(url: str) -> str:
     if len(protocols) > 1 and "web.archive.org" not in url:
         LOGGER.debug("double url: %s %s", len(protocols), url)
         match = SELECTION.match(url)
-        if match and validate_url(match[1])[0] is True:
+        if match and is_valid_url(match[1]):
             url = match[1]
             LOGGER.debug("taking url: %s", url)
         else:
             match = MIDDLE_URL.match(url)
-            if match and validate_url(match[1])[0] is True:
+            if match and is_valid_url(match[1]):
                 url = match[1]
                 LOGGER.debug("taking url: %s", url)
     # too long and garbled URLs e.g. due to quotes URLs
     # https://github.com/cocrawler/cocrawler/blob/main/cocrawler/urls.py
-    if len(url) > 500:  # arbitrary choice
-        match = TRAILING_PARTS.match(url)
-        if match:
-            url = match[1]
+    # if len(url) > 500:  # arbitrary choice
+    match = TRAILING_PARTS.match(url)
+    if match:
+        url = match[1]
     if len(url) > 500:
         LOGGER.debug("invalid-looking link %s of length %d", url[:50] + "…", len(url))
 
@@ -97,20 +109,20 @@ def scrub_url(url: str) -> str:
 
 
 def clean_query(
-    parsed_url: ParseResult, strict: bool = False, language: Optional[str] = None
-) -> ParseResult:
-    """Strip unwanted query elements"""
-    if len(parsed_url.query) > 0:
-        qdict = parse_qs(parsed_url.query)
+    querystring: str, strict: bool = False, language: Optional[str] = None
+) -> str:
+    "Strip unwanted query elements"
+    if querystring:
+        qdict = parse_qs(querystring)
         newqdict = {}
-        for qelem in sorted(qdict.keys()):
+        for qelem in sorted(qdict):
             teststr = qelem.lower()
             # control param
-            if (
-                strict
-                and teststr not in ALLOWED_PARAMS
-                and teststr not in CONTROL_PARAMS
-            ):
+            if strict:
+                if teststr not in ALLOWED_PARAMS and teststr not in CONTROL_PARAMS:
+                    continue
+            # get rid of trackers
+            elif TRACKERS_RE.search(teststr):
                 continue
             # control language
             if language is not None and teststr in CONTROL_PARAMS:
@@ -124,38 +136,71 @@ def clean_query(
                     raise ValueError
             # insert
             newqdict[qelem] = qdict[qelem]
-        newstring = urlencode(newqdict, doseq=True)
-        parsed_url = parsed_url._replace(query=newstring)
-    return parsed_url
+        return urlencode(newqdict, doseq=True)
+    return querystring
+
+
+def decode_punycode(string: str) -> str:
+    "Probe for punycode in lower-cased hostname and try to decode it."
+    if "xn--" not in string:
+        return string
+
+    parts = []
+
+    for part in string.split("."):
+        if part.lower().startswith("xn--"):
+            try:
+                part = part.encode("utf8").decode("idna")
+            except UnicodeError:
+                LOGGER.debug("invalid utf/idna string: %s", part)
+        parts.append(part)
+
+    return ".".join(parts)
+
+
+def normalize_part(url_part: str) -> str:
+    """Normalize URLs parts (specifically path and fragment) while
+    accounting for certain characters."""
+    # "~" for compatibility with Python 3.6
+    return quote(url_part, safe="/%!=:,-~")
+
+
+def normalize_fragment(fragment: str, language: Optional[str] = None) -> str:
+    "Look for trackers in URL fragments using query analysis, normalize the output."
+    if "=" in fragment:
+        if "&" in fragment:
+            fragment = clean_query(fragment, False, language)
+        elif TRACKERS_RE.search(fragment):
+            fragment = ""
+    return normalize_part(fragment)
 
 
 def normalize_url(
-    parsed_url: Union[ParseResult, str],
+    parsed_url: Union[SplitResult, str],
     strict: bool = False,
     language: Optional[str] = None,
 ) -> str:
-    """Takes a URL string or a parsed URL and returns a (basically) normalized URL string"""
-    if not isinstance(parsed_url, ParseResult):
-        parsed_url = urlparse(parsed_url)
-    # elif not isinstance(parsed_url, str):
-    #    raise TypeError('wrong input type: %s', type(parsed_url))
+    "Takes a URL string or a parsed URL and returns a normalized URL string"
+    parsed_url = _parse(parsed_url)
+    # lowercase + remove fragments + normalize punycode
+    scheme = parsed_url.scheme.lower()
+    netloc = parsed_url.netloc.lower()
     # port
-    if parsed_url.port is not None and parsed_url.port in (80, 443):
-        parsed_url = parsed_url._replace(netloc=NETLOC_RE.sub("", parsed_url.netloc))
+    try:
+        if parsed_url.port and parsed_url.port in (80, 443):
+            netloc = NETLOC_RE.sub("", netloc)
+    except ValueError:
+        pass  # Port could not be cast to integer value
+    # lowercase + remove fragments + normalize punycode
+    netloc = decode_punycode(netloc)
     # path: https://github.com/saintamh/alcazar/blob/master/alcazar/utils/urls.py
-    newpath = PATH1.sub("/", parsed_url.path)
-    # Leading /../'s in the path are removed
-    newpath = PATH2.sub("", newpath)
-    # fragment
-    newfragment = "" if strict else parsed_url.fragment
-    # lowercase + remove fragments
-    parsed_url = parsed_url._replace(
-        scheme=parsed_url.scheme.lower(),
-        netloc=parsed_url.netloc.lower(),
-        path=newpath,
-        fragment=newfragment,
-    )
+    # leading /../'s in the path are removed
+    newpath = normalize_part(PATH2.sub("", PATH1.sub("/", parsed_url.path)))
     # strip unwanted query elements
-    parsed_url = clean_query(parsed_url, strict, language)
+    newquery = clean_query(parsed_url.query, strict, language) or ""
+    if newquery and newpath == "":
+        newpath = "/"
+    # fragment
+    newfragment = "" if strict else normalize_fragment(parsed_url.fragment, language)
     # rebuild
-    return parsed_url.geturl()
+    return urlunsplit([scheme, netloc, newpath, newquery, newfragment])
