@@ -9,14 +9,16 @@ Core functions needed to make the module work.
 import logging
 import re
 
-# from functools import cmp_to_key
-from random import sample
 from typing import List, Optional, Set, Tuple
+from urllib.robotparser import RobotFileParser
 
 from .clean import normalize_url, scrub_url
 from .filters import (
     basic_filter,
+    domain_filter,
     extension_filter,
+    is_navigation_page,
+    is_not_crawlable,
     lang_filter,
     path_filter,
     type_filter,
@@ -24,15 +26,20 @@ from .filters import (
 )
 from .network import redirection_test
 from .settings import BLACKLIST
-from .urlstore import UrlStore
-from .urlutils import extract_domain, fix_relative_urls, is_external, is_known_link
+from .urlutils import (
+    extract_domain,
+    get_base_url,
+    fix_relative_urls,
+    is_external,
+    is_known_link,
+)
 
 
 LOGGER = logging.getLogger(__name__)
 
 FIND_LINKS_REGEX = re.compile(r"<a [^<>]+?>", re.I)
 HREFLANG_REGEX = re.compile(r'hreflang=["\']?([a-z-]+)', re.I)
-LINK_REGEX = re.compile(r'href=["\']?([^ ]+?)(["\']|[ >])', re.I)
+LINK_REGEX = re.compile(r'href=["\']?([^ ]+?)(["\' >])', re.I)
 
 
 def check_url(
@@ -93,6 +100,11 @@ def check_url(
             LOGGER.debug("rejected, extension filter: %s", url)
             raise ValueError
 
+        # unsuitable domain/host name
+        if domain_filter(parsed_url.netloc) is False:
+            LOGGER.debug("rejected, domain name: %s", url)
+            raise ValueError
+
         # strict content filtering
         if strict and path_filter(parsed_url.path, parsed_url.query) is False:
             LOGGER.debug("rejected, path filter: %s", url)
@@ -118,66 +130,12 @@ def check_url(
     return url, domain
 
 
-def sample_urls(
-    input_urls: List[str],
-    samplesize: int,
-    exclude_min: Optional[int] = None,
-    exclude_max: Optional[int] = None,
-    strict: bool = False,
-    verbose: bool = False,
-) -> List[str]:
-    """Sample a list of URLs by domain name, optionally using constraints on their number"""
-    # logging
-    if verbose:
-        LOGGER.setLevel(logging.DEBUG)
-    else:
-        LOGGER.setLevel(logging.ERROR)
-    # deduplicate
-    input_urls = list(dict.fromkeys(input_urls))
-    # validate
-    input_urls = [
-        u
-        for u in input_urls
-        if check_url(u, strict=strict, with_redirects=False) is not None
-    ]
-    # store
-    output_urls = []
-    urlstore = UrlStore(compressed=False, language=None, strict=strict)
-    urlstore.add_urls(input_urls)
-    # iterate
-    for domain in urlstore.urldict:  # key=cmp_to_key(locale.strcoll)
-        urlpaths = [p.urlpath for p in urlstore._load_urls(domain)]
-        # too few or too many URLs
-        if (
-            exclude_min is not None
-            and len(urlpaths) < exclude_min
-            or exclude_max is not None
-            and len(urlpaths) > exclude_max
-        ):
-            LOGGER.info("discarded (size): %s\t\turls: %s", domain, len(urlpaths))
-            continue
-        # copy all URLs
-        if len(urlpaths) <= samplesize:
-            output_urls.extend([domain + p for p in urlpaths])
-            LOGGER.info("%s\t\turls: %s", domain, len(urlpaths))
-            continue
-        # sample
-        mysample = sorted(sample(urlpaths, k=samplesize))
-        output_urls.extend([domain + p for p in mysample])
-        LOGGER.info(
-            "%s\t\turls: %s\tprop.: %s",
-            domain,
-            len(mysample),
-            len(mysample) / len(urlpaths),
-        )
-    # return gathered URLs
-    return output_urls
-
-
 def extract_links(
     pagecontent: str,
-    base_url: str,
-    external_bool: bool,
+    url: Optional[str] = None,
+    base_url: Optional[str] = None,
+    external_bool: bool = False,
+    no_filter: bool = False,
     language: Optional[str] = None,
     strict: bool = True,
     with_nav: bool = False,
@@ -187,9 +145,11 @@ def extract_links(
     """Filter links in a HTML document using a series of heuristics
     Args:
         pagecontent: whole page in binary format
-        base_url: beginning of the URL, without path, fragment and query
-        external: set to True for external links only, False for
+        url: full URL of the original page
+        base_url: deprecated, legacy only
+        external_bool: set to True for external links only, False for
                   internal links only
+        no_filter: override settings and bypass checks to return all possible URLs
         language: set target language (ISO 639-1 codes)
         strict: set to True for stricter filtering
         with_nav: set to True to include navigation pages instead of discarding them
@@ -202,17 +162,19 @@ def extract_links(
     Raises:
         Nothing.
     """
+    base_url = base_url or get_base_url(url)
+    url = url or base_url
     candidates, validlinks = set(), set()  # type: Set[str], Set[str]
-    if pagecontent is None or not pagecontent:
+    if not pagecontent:
         return validlinks
     # define host reference
-    if reference is None:
-        reference = base_url
+    reference = reference or base_url
     # extract links
-    for match in FIND_LINKS_REGEX.finditer(pagecontent):
-        link = match[0]
+    for link in (m[0] for m in FIND_LINKS_REGEX.finditer(pagecontent)):
+        if "rel" in link and "nofollow" in link:
+            continue
         # https://en.wikipedia.org/wiki/Hreflang
-        if language is not None and "hreflang" in link:
+        if no_filter is False and language is not None and "hreflang" in link:
             langmatch = HREFLANG_REGEX.search(link)
             if langmatch and (
                 langmatch[1].startswith(language) or langmatch[1] == "x-default"
@@ -229,24 +191,61 @@ def extract_links(
     for link in candidates:
         # repair using base
         if not link.startswith("http"):
-            link = fix_relative_urls(base_url, link)
+            link = fix_relative_urls(url, link)
         # check
-        checked = check_url(
-            link,
-            strict=strict,
-            with_nav=with_nav,
-            with_redirects=redirects,
-            language=language,
-        )
-        if checked is None:
-            continue
-        link = checked[0]
-        # external/internal links
-        if external_bool != is_external(link, reference):
-            continue
+        if no_filter is False:
+            checked = check_url(
+                link,
+                strict=strict,
+                with_nav=with_nav,
+                with_redirects=redirects,
+                language=language,
+            )
+            if checked is None:
+                continue
+            link = checked[0]
+            # external/internal links
+            if external_bool != is_external(
+                url=link, reference=reference, ignore_suffix=True
+            ):
+                continue
         if is_known_link(link, validlinks):
             continue
         validlinks.add(link)
     # return
     LOGGER.info("%s links found – %s valid links", len(candidates), len(validlinks))
     return validlinks
+
+
+def filter_links(
+    htmlstring: str,
+    url: Optional[str],
+    base_url: Optional[str] = None,
+    lang: Optional[str] = None,
+    rules: Optional[RobotFileParser] = None,
+    external: bool = False,
+    strict: bool = False,
+    with_nav: bool = True,
+) -> Tuple[List[str], List[str]]:
+    "Find links in a HTML document, filter them and add them to the data store."
+    links, links_priority = [], []
+    url = url or base_url
+    for link in extract_links(
+        pagecontent=htmlstring,
+        url=url,
+        external_bool=external,
+        language=lang,
+        strict=strict,
+        with_nav=with_nav,
+    ):
+        # sanity check
+        if is_not_crawlable(link) or (
+            rules is not None and not rules.can_fetch("*", link)
+        ):
+            continue
+        # store
+        if is_navigation_page(link):
+            links_priority.append(link)
+        else:
+            links.append(link)
+    return links, links_priority
