@@ -9,8 +9,9 @@ Bundles functions needed to target text content and validate the input.
 import logging
 import re
 
+from ipaddress import ip_address
 from typing import Any, Optional, Tuple
-from urllib.parse import urlparse
+from urllib.parse import urlsplit
 
 from langcodes import Language, tag_is_valid
 
@@ -19,22 +20,65 @@ from .langinfo import COUNTRY_CODES, LANGUAGE_CODES
 
 LOGGER = logging.getLogger(__name__)
 
+
+# domain/host names
+IP_SET = {
+    ".",
+    ":",
+    "0",
+    "1",
+    "2",
+    "3",
+    "4",
+    "5",
+    "6",
+    "7",
+    "8",
+    "9",
+    "a",
+    "b",
+    "c",
+    "d",
+    "e",
+    "f",
+}
+
+# https://github.com/python-validators/validators/blob/master/src/validators/domain.py
+VALID_DOMAIN = re.compile(
+    # First character of the domain
+    r"^(?:[a-zA-Z0-9]"
+    # Sub domain + hostname
+    + r"(?:[a-zA-Z0-9-_]{0,61}[A-Za-z0-9])?\.)"
+    # First 61 characters of the gTLD
+    + r"+[A-Za-z0-9][A-Za-z0-9-_]{0,61}"
+    # Last character of the gTLD
+    + r"[A-Za-z]$",
+    re.IGNORECASE,
+)
+
+UNSUITABLE_DOMAIN = re.compile(r"[0-9]+\.")
+
 # content filters
-WORDPRESS_CONTENT = re.compile(
-    r"/(?:page|seite|user|search|gallery|gall?erie|labels|archives|uploads|modules|attachment)/|/(?:tags?|schlagwort|category|cat|kategorie|kat|auth?or)/[^/]+/?$",
+SITE_STRUCTURE = re.compile(
+    # wordpress
+    r"/(?:paged?|seite|search|suche|gall?er[a-z]{1,2}|labels|archives|uploads|modules|attachment|wp-admin|wp-content|wp-includes|wp-json|wp-themes|oembed)/|"
+    # wordpress + short URL
+    r"[/_-](?:tags?|schlagwort|[ck]ategor[a-z]{1,2}|[ck]at|auth?or|user)/[^/]+/?$|"
+    # mixed/blogspot
+    r"[^0-9]/[0-9]+/[0-9]+/$|[^0-9]/[0-9]{4}/$|"
+    # blogspot
+    r"_archive\.html$",
     re.IGNORECASE,
 )
 FILE_TYPE = re.compile(
-    r"\.(atom|json|css|xml|js|jpg|jpeg|png|gif|tiff|pdf|ogg|mp3|m4a|aac|avi|mp4|mov|webm|flv|ico|pls|zip|tar|gz|iso|swf)\b",
+    r"\.(atom|json|css|xml|js|jpg|jpeg|png|svg|gif|tiff|pdf|ogg|mp3|m4a|aac|avi|mp4|mov|web[mp]|flv|ico|pls|zip|tar|gz|iso|swf|woff|eot|ttf)\b|"
+    r"[/-](img|jpg|png)(\b|_)",
     re.IGNORECASE,
 )  # (?=[&?])
-UNDESIRABLE = re.compile(
-    r"\b(?:add?s?|banner|doubleclick|livestream|tradedoubler)\b|/oembed\b"
-)
 ADULT_AND_VIDEOS = re.compile(
-    r"\b(?:live|videos?|adult|amateur|arsch|cams?|cash|fick|gangbang|incest|porn|sexyeroti[ck]|sexcam|swinger|xxx|bild\-?kontakte)\b",
+    r"[/_-](?:bild-?kontakte|fick|gangbang|incest|live-?cams?|live-?chat|porno?|sexcam|sexyeroti[ck]|swinger|x{3})\b",
     re.IGNORECASE,
-)  # ass|orgasm ?
+)
 
 # language filter
 PATH_LANG_FILTER = re.compile(
@@ -47,20 +91,23 @@ HOST_LANG_FILTER = re.compile(
 
 # navigation/crawls
 NAVIGATION_FILTER = re.compile(
-    r"/(archives|auth?or|cat|category|kat|kategorie|page|schlagwort|seite|tags?|topics?|user)/|\?p=[0-9]+",
+    r"[/_-](archives|auth?or|[ck]at|category|kategorie|paged?|schlagwort|seite|tags?|topics?|user)/|\?p=[0-9]+",
     re.IGNORECASE,
 )
 NOTCRAWLABLE = re.compile(
-    r"/(login|impressum|imprint)(\.[a-z]{3,4})?/?$|/login\?|/(javascript:|mailto:|tel\.?:|whatsapp:)",
+    r"/([ck]onta[ck]t|datenschutzerkl.{1,2}rung|login|impressum|imprint)(\.[a-z]{3,4})?/?$|/login\?|"
+    r"/(javascript:|mailto:|tel\.?:|whatsapp:)",
     re.IGNORECASE,
 )
 # |/(www\.)?(facebook\.com|google\.com|instagram\.com|twitter\.com)/
-INDEX_PAGE_FILTER = re.compile(r".{0,5}/index(\.[a-z]{3,4})?/?$", re.IGNORECASE)
+INDEX_PAGE_FILTER = re.compile(
+    r".{0,5}/(default|home|index)(\.[a-z]{3,5})?/?$", re.IGNORECASE
+)
 
 # document types
 EXTENSION_REGEX = re.compile(r"\.[a-z]{2,5}$")
 # https://en.wikipedia.org/wiki/List_of_file_formats#Web_page
-WHITELISTED_EXTENSIONS = (
+WHITELISTED_EXTENSIONS = {
     ".adp",
     ".amp",
     ".asp",
@@ -69,7 +116,7 @@ WHITELISTED_EXTENSIONS = (
     ".cgi",
     ".do",
     ".htm",
-    "html",
+    ".html",
     ".htx",
     ".jsp",
     ".mht",
@@ -85,7 +132,7 @@ WHITELISTED_EXTENSIONS = (
     ".txt",
     ".xhtml",
     ".xml",
-)
+}
 
 # territories whitelist
 # see also: https://babel.pocoo.org/en/latest/api/languages.html
@@ -102,11 +149,41 @@ def basic_filter(url: str) -> bool:
     return bool(url.startswith("http") and 10 <= len(url) < 500)
 
 
+def domain_filter(domain: str) -> bool:
+    "Find invalid domain/host names"
+    # IPv4 or IPv6
+    if not set(domain).difference(IP_SET):
+        try:
+            ip_address(domain)
+        except ValueError:
+            return False
+        return True
+
+    # malformed domains
+    try:
+        if not VALID_DOMAIN.match(domain.encode("idna").decode("utf-8")):
+            return False
+    except UnicodeError:
+        return False
+
+    # unsuitable content
+    if UNSUITABLE_DOMAIN.match(domain):
+        return False
+
+    if FILE_TYPE.search(domain):
+        return False
+
+    extension_match = EXTENSION_REGEX.search(domain)
+    if extension_match and extension_match[0] in WHITELISTED_EXTENSIONS:
+        return False
+
+    return True
+
+
 def extension_filter(urlpath: str) -> bool:
     """Filter based on file extension"""
-    return bool(
-        not EXTENSION_REGEX.search(urlpath) or urlpath.endswith(WHITELISTED_EXTENSIONS)
-    )
+    extension_match = EXTENSION_REGEX.search(urlpath)
+    return not extension_match or extension_match[0] in WHITELISTED_EXTENSIONS
 
 
 def langcodes_score(language: str, segment: str, score: int) -> int:
@@ -167,10 +244,7 @@ def path_filter(urlpath: str, query: str) -> bool:
     """Filters based on URL path: index page, imprint, etc."""
     if NOTCRAWLABLE.search(urlpath):
         return False
-    if INDEX_PAGE_FILTER.match(urlpath) and len(query) == 0:
-        # print('#', urlpath, INDEX_PAGE_FILTER.match(urlpath), query)
-        return False
-    return True
+    return bool(not INDEX_PAGE_FILTER.match(urlpath) or query)
 
 
 def type_filter(url: str, strict: bool = False, with_nav: bool = False) -> bool:
@@ -180,15 +254,10 @@ def type_filter(url: str, strict: bool = False, with_nav: bool = False) -> bool:
         # feeds
         if url.endswith(("/feed", "/rss")):
             raise ValueError
-        # wordpress website structure
-        if WORDPRESS_CONTENT.search(url) and (
-            not with_nav or not is_navigation_page(url)
-        ):
+        # website structure
+        if SITE_STRUCTURE.search(url) and (not with_nav or not is_navigation_page(url)):
             raise ValueError
-        # not suitable: ads, adult and embedded content
-        if UNDESIRABLE.search(url):
-            raise ValueError
-        # type hidden in parameters + video content
+        # type (also hidden in parameters), videos, adult content
         if strict and (FILE_TYPE.search(url) or ADULT_AND_VIDEOS.search(url)):
             raise ValueError
     except ValueError:
@@ -200,7 +269,7 @@ def type_filter(url: str, strict: bool = False, with_nav: bool = False) -> bool:
 def validate_url(url: Optional[str]) -> Tuple[bool, Any]:
     """Parse and validate the input"""
     try:
-        parsed_url = urlparse(url)
+        parsed_url = urlsplit(url)
     except ValueError:
         return False, None
     if not bool(parsed_url.scheme) or parsed_url.scheme not in (
@@ -219,6 +288,11 @@ def validate_url(url: Optional[str]) -> Tuple[bool, Any]:
     return True, parsed_url
 
 
+def is_valid_url(url: Optional[str]) -> bool:
+    "Determine if a given string is a valid URL."
+    return validate_url(url)[0]
+
+
 def is_navigation_page(url: str) -> bool:
     """Determine if the URL is related to navigation and overview pages
     rather than content pages, e.g. /page/1 vs. article page."""
