@@ -7,17 +7,41 @@ Functions devoted to requests over the WWW.
 
 import logging
 
-import urllib3  # type: ignore
+import urllib3
 
 
 LOGGER = logging.getLogger(__name__)
+urllib3.disable_warnings()
+
 
 RETRY_STRATEGY = urllib3.util.Retry(
-    total=5,
-    redirect=5,
+    total=2,
+    redirect=2,
     raise_on_redirect=False,
+    status_forcelist=[
+        429,
+        499,
+        500,
+        502,
+        503,
+        504,
+        509,
+        520,
+        521,
+        522,
+        523,
+        524,
+        525,
+        526,
+        527,
+        530,
+        598,
+    ],  # unofficial: https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#Unofficial_codes
+    backoff_factor=1,
+)
+HTTP_POOL = urllib3.PoolManager(
+    cert_reqs="CERT_NONE", num_pools=100, retries=RETRY_STRATEGY, timeout=10
 )
-HTTP_POOL = urllib3.PoolManager(retries=RETRY_STRATEGY)
 
 ACCEPTABLE_CODES = {200, 300, 301, 302, 303, 304, 305, 306, 307, 308}
 
@@ -38,7 +62,7 @@ def redirection_test(url: str) -> str:
     #    "User-Agent" : str(sample(settings.USER_AGENTS, 1)), # select a random user agent
     # })
     try:
-        rhead = HTTP_POOL.request("HEAD", url)
+        rhead = HTTP_POOL.request("HEAD", url)  # type:ignore[no-untyped-call]
     except Exception as err:
         LOGGER.exception("unknown error: %s %s", url, err)
     else:
@@ -47,4 +71,4 @@ def redirection_test(url: str) -> str:
             LOGGER.debug("result found: %s %s", rhead.geturl(), rhead.status)
             return rhead.geturl()  # type: ignore
     # else:
-    raise ValueError("cannot reach URL: %s", url)
+    raise ValueError(f"cannot reach URL: ${url}")
