@@ -1,37 +1,60 @@
 """
-Defines a URL store which holds URLs along with relevant information.
+Defines a URL store which holds URLs along with relevant information and entails crawling helpers.
 """
 
 import bz2
+import gc
 import logging
 import pickle
 import signal
 import sys
+import zlib
 
 from collections import defaultdict, deque
 from datetime import datetime, timedelta
+from enum import Enum
 from threading import Lock
-from typing import Any, DefaultDict, Deque, Dict, List, Optional, Tuple, Union
+from typing import (
+    Any,
+    DefaultDict,
+    Deque,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
 from urllib.robotparser import RobotFileParser
 
+from .clean import normalize_url
+from .core import filter_links
 from .filters import lang_filter, validate_url
-from .urlutils import get_host_and_path, is_known_link
+from .meta import clear_caches
+from .urlutils import get_base_url, get_host_and_path, is_known_link
 
 
 LOGGER = logging.getLogger(__name__)
 
 
+class State(Enum):
+    "Record state information about a domain or host."
+    OPEN = 1
+    ALL_VISITED = 2
+    BUSTED = 3
+
+
 class DomainEntry:
     "Class to record host-related information and URL paths."
-    __slots__ = ("all_visited", "count", "rules", "timestamp", "tuples")
+    __slots__ = ("count", "rules", "state", "timestamp", "total", "tuples")
 
     def __init__(self) -> None:
-        self.all_visited: bool = False
         self.count: int = 0
         self.rules: Optional[RobotFileParser] = None
-        self.tuples: Deque[UrlPathTuple] = deque()
+        self.state: State = State.OPEN
         self.timestamp: Optional[Any] = None
+        self.total: int = 0
+        self.tuples: Deque[UrlPathTuple] = deque()
 
 
 class UrlPathTuple:
@@ -52,6 +75,7 @@ class UrlStore:
         compressed: bool = False,
         language: Optional[str] = None,
         strict: bool = False,
+        verbose: bool = False,
     ) -> None:
         self.compressed: bool = compressed
         self.done: bool = False
@@ -61,38 +85,23 @@ class UrlStore:
         self._lock: Lock = Lock()
 
         def dump_unvisited_urls(num: Any, frame: Any) -> None:
-            LOGGER.warning(
+            LOGGER.debug(
                 "Processing interrupted, dumping unvisited URLs from %s hosts",
                 len(self.urldict),
             )
-            for domain in self.urldict:
-                print(
-                    "\n".join(
-                        [
-                            domain + u.urlpath
-                            for u in self._load_urls(domain)
-                            if not u.visited
-                        ]
-                    ),
-                    file=sys.stderr,
-                )
-            sys.exit()
+            self.print_unvisited_urls()
+            sys.exit(1)
 
         # don't use the following on Windows
-        if not sys.platform.startswith("win"):
+        if verbose and not sys.platform.startswith("win"):
             signal.signal(signal.SIGINT, dump_unvisited_urls)
             signal.signal(signal.SIGTERM, dump_unvisited_urls)
 
-    # def _filter_urlpaths(self, domain, urls):
-    #    if self.validation or self.language is not None:
-    #        return [u for u in urls if self._filter_url(domain + u)]
-    #    return urls
-
     def _buffer_urls(
         self, data: List[str], visited: bool = False
     ) -> DefaultDict[str, Deque[UrlPathTuple]]:
         inputdict: DefaultDict[str, Deque[UrlPathTuple]] = defaultdict(deque)
-        for url in list(dict.fromkeys(data)):
+        for url in dict.fromkeys(data):
             # segment URL and add to domain dictionary
             try:
                 # validate
@@ -107,6 +116,9 @@ class UrlStore:
                 ):
                     LOGGER.debug("Wrong language: %s", url)
                     raise ValueError
+                parsed_url = normalize_url(
+                    parsed_url, strict=self.strict, language=self.language
+                )
                 hostinfo, urlpath = get_host_and_path(parsed_url)
                 inputdict[hostinfo].append(UrlPathTuple(urlpath, visited))
             except (TypeError, ValueError):
@@ -114,10 +126,16 @@ class UrlStore:
         return inputdict
 
     def _load_urls(self, domain: str) -> Deque[UrlPathTuple]:
-        value = self.urldict[domain].tuples
-        if isinstance(value, bytes):
-            return pickle.loads(bz2.decompress(value))
-        return value
+        if domain in self.urldict:
+            if self.compressed:
+                return pickle.loads(bz2.decompress(self.urldict[domain].tuples))  # type: ignore
+            return self.urldict[domain].tuples
+        return deque()
+
+    def _set_done(self) -> None:
+        if not self.done and all(self.is_exhausted_domain(d) for d in self.urldict):
+            with self._lock:
+                self.done = True
 
     def _store_urls(
         self,
@@ -126,32 +144,55 @@ class UrlStore:
         timestamp: Optional[datetime] = None,
         to_left: Optional[Deque[UrlPathTuple]] = None,
     ) -> None:
+        # http/https switch
+        if domain.startswith("http://"):
+            candidate = "https" + domain[4:]
+            # switch
+            if candidate in self.urldict:
+                domain = candidate
+        elif domain.startswith("https://"):
+            candidate = "http" + domain[5:]
+            # replace entry
+            if candidate in self.urldict:
+                self.urldict[domain] = self.urldict[candidate]
+                del self.urldict[candidate]
+
+        # load URLs or create entry
         if domain in self.urldict:
+            # discard if busted
+            if self.urldict[domain].state is State.BUSTED:
+                return
             urls = self._load_urls(domain)
             known = {u.urlpath for u in urls}
         else:
             urls = deque()
             known = set()
+
         # check if the link or its variants are known
         if to_right is not None:
             urls.extend(t for t in to_right if not is_known_link(t.urlpath, known))
         if to_left is not None:
             urls.extendleft(t for t in to_left if not is_known_link(t.urlpath, known))
-        # use lock
+
         with self._lock:
-            # compression
             if self.compressed:
-                self.urldict[domain].tuples = bz2.compress(  # type: ignore
+                self.urldict[domain].tuples = bz2.compress(  # type: ignore[assignment]
                     pickle.dumps(urls, protocol=4)
                 )
             else:
                 self.urldict[domain].tuples = urls
-            # adjust all_visited status
-            self.urldict[domain].all_visited = all(u.visited for u in urls)
-            # timestamp/backoff value
+            self.urldict[domain].total = len(urls)
+
             if timestamp is not None:
                 self.urldict[domain].timestamp = timestamp
 
+            if all(u.visited for u in urls):
+                self.urldict[domain].state = State.ALL_VISITED
+            else:
+                self.urldict[domain].state = State.OPEN
+                if self.done:
+                    self.done = False
+
     def _search_urls(
         self, urls: List[str], switch: Optional[int] = None
     ) -> List[Union[Any, str]]:
@@ -165,12 +206,7 @@ class UrlStore:
             # examine domain
             if hostinfo != last_domain:
                 last_domain = hostinfo
-                if switch == 1:
-                    known_paths = {u.urlpath: None for u in self._load_urls(hostinfo)}
-                elif switch == 2:
-                    known_paths = {
-                        u.urlpath: u.visited for u in self._load_urls(hostinfo)
-                    }
+                known_paths = {u.urlpath: u.visited for u in self._load_urls(hostinfo)}
             # run checks: case 1: the path matches, case 2: visited URL
             if urlpath in known_paths and (
                 switch == 1 or (switch == 2 and known_paths[urlpath])
@@ -182,7 +218,7 @@ class UrlStore:
     def _timestamp(self, domain: str) -> Optional[datetime]:
         return self.urldict[domain].timestamp
 
-    # URL MANIPULATION AND INFO
+    # ADDITIONS AND DELETIONS
 
     def add_urls(
         self,
@@ -200,83 +236,133 @@ class UrlStore:
             for host, urltuples in self._buffer_urls(appendleft, visited).items():
                 self._store_urls(host, to_left=urltuples)
 
-    def is_known(self, url: str) -> bool:
-        "Check if the given URL has already been stored."
-        hostinfo, urlpath = get_host_and_path(url)
-        values = self._load_urls(hostinfo)
-        # returns False if domain or URL is new
-        return urlpath in {u.urlpath for u in values}
+    def add_from_html(
+        self,
+        htmlstring: str,
+        url: str,
+        external: bool = False,
+        lang: Optional[str] = None,
+        with_nav: bool = True,
+    ) -> None:
+        "Find links in a HTML document, filter them and add them to the data store."
+        # lang = lang or self.language
+        base_url = get_base_url(url)
+        rules = self.get_rules(base_url)
+        links, links_priority = filter_links(
+            htmlstring=htmlstring,
+            url=url,
+            external=external,
+            lang=lang or self.language,
+            rules=rules,
+            strict=self.strict,
+            with_nav=with_nav,
+        )
+        self.add_urls(urls=links, appendleft=links_priority)
 
-    def find_known_urls(self, domain: str) -> List[str]:
-        """Get all already known URLs for the given domain (ex. "https://example.org")."""
-        values = self._load_urls(domain)
-        return [domain + u.urlpath for u in values]
+    def discard(self, domains: List[str]) -> None:
+        "Declare domains void and prune the store."
+        with self._lock:
+            for d in domains:
+                self.urldict[d] = DomainEntry()
+                self.urldict[d].state = State.BUSTED
+        self._set_done()
+        num = gc.collect()
+        LOGGER.debug("%s objects in GC after UrlStore.discard", num)
 
-    def filter_unknown_urls(self, urls: List[str]) -> List[str]:
-        "Take a list of URLs and return the currently unknown ones."
-        return self._search_urls(urls, switch=1)
+    def reset(self) -> None:
+        "Re-initialize the URL store."
+        with self._lock:
+            self.urldict = defaultdict(DomainEntry)
+        clear_caches()
+        num = gc.collect()
+        LOGGER.debug("UrlStore reset, %s objects in GC", num)
+
+    # DOMAINS / HOSTNAMES
 
     def get_known_domains(self) -> List[str]:
         "Return all known domains as a list."
         return list(self.urldict)
 
+    def get_unvisited_domains(self) -> List[str]:
+        """Find all domains for which there are unvisited URLs
+        and potentially adjust done meta-information."""
+        unvisited = []
+        if not self.done:
+            unvisited = [d for d in self.urldict if not self.is_exhausted_domain(d)]
+            if not unvisited:
+                self._set_done()
+        return unvisited
+
     def is_exhausted_domain(self, domain: str) -> bool:
         "Tell if all known URLs for the website have been visited."
         if domain in self.urldict:
-            return self.urldict[domain].all_visited
-        raise KeyError("website not in store")
+            return self.urldict[domain].state in (State.ALL_VISITED, State.BUSTED)
+        return False
+        # raise KeyError("website not in store")
+
+    def unvisited_websites_number(self) -> int:
+        "Return the number of websites for which there are still URLs to visit."
+        return len(self.get_unvisited_domains())
 
     # URL-BASED QUERIES
 
-    def has_been_visited(self, url: str) -> bool:
-        "Check if the given URL has already been visited.."
-        hostinfo, urlpath = get_host_and_path(url)
-        values = self._load_urls(hostinfo)
-        known_urlpaths = {u.urlpath: u.visited for u in values}
-        # defaults to None, thus False
-        return known_urlpaths.get(urlpath) or False
+    def find_known_urls(self, domain: str) -> List[str]:
+        """Get all already known URLs for the given domain (ex. "https://example.org")."""
+        return [domain + u.urlpath for u in self._load_urls(domain)]
 
     def find_unvisited_urls(self, domain: str) -> List[str]:
         "Get all unvisited URLs for the given domain."
-        values = self._load_urls(domain)
-        return [domain + u.urlpath for u in values if not u.visited]
+        if not self.is_exhausted_domain(domain):
+            return [
+                domain + u.urlpath for u in self._load_urls(domain) if not u.visited
+            ]
+        return []
+
+    def filter_unknown_urls(self, urls: List[str]) -> List[str]:
+        "Take a list of URLs and return the currently unknown ones."
+        return self._search_urls(urls, switch=1)
 
     def filter_unvisited_urls(self, urls: List[str]) -> List[Union[Any, str]]:
         "Take a list of URLs and return the currently unvisited ones."
         return self._search_urls(urls, switch=2)
 
-    def unvisited_websites_number(self) -> int:
-        "Return the number of websites for which there are still URLs to visit."
-        return len([d for d in self.urldict if not self.urldict[d].all_visited])
+    def has_been_visited(self, url: str) -> bool:
+        "Check if the given URL has already been visited."
+        return not bool(self.filter_unvisited_urls([url]))
+
+    def is_known(self, url: str) -> bool:
+        "Check if the given URL has already been stored."
+        hostinfo, urlpath = get_host_and_path(url)
+        # returns False if domain or URL is new
+        return urlpath in {u.urlpath for u in self._load_urls(hostinfo)}
 
     # DOWNLOADS
 
-    def get_url(self, domain: str) -> Optional[str]:
+    def get_url(self, domain: str, as_visited: bool = True) -> Optional[str]:
         "Retrieve a single URL and consider it to be visited (with corresponding timestamp)."
         # not fully used
-        if not self.urldict[domain].all_visited:
+        if not self.is_exhausted_domain(domain):
             url_tuples = self._load_urls(domain)
             # get first non-seen url
             for url in url_tuples:
                 if not url.visited:
-                    url.visited = True
-                    with self._lock:
-                        self.urldict[domain].count += 1
-                    self._store_urls(domain, url_tuples, timestamp=datetime.now())
+                    # store information
+                    if as_visited:
+                        url.visited = True
+                        with self._lock:
+                            self.urldict[domain].count += 1
+                        self._store_urls(domain, url_tuples, timestamp=datetime.now())
                     return domain + url.urlpath
         # nothing to draw from
         with self._lock:
-            self.urldict[domain].all_visited = True
+            self.urldict[domain].state = State.ALL_VISITED
+        self._set_done()
         return None
 
     def get_download_urls(self, timelimit: int = 10) -> Optional[List[str]]:
         """Get a list of immediately downloadable URLs according to the given
         time limit per domain."""
-        with self._lock:
-            potential = [d for d in self.urldict if not self.urldict[d].all_visited]
-        if not potential:
-            self.done = True
-            return None
+        potential = self.get_unvisited_domains()
         targets = []
         for domain in potential:
             timestamp = self._timestamp(domain)
@@ -294,18 +380,16 @@ class UrlStore:
         """Get up to the specified number of URLs along with a suitable
         backoff schedule (in seconds)."""
         # see which domains are free
-        with self._lock:
-            potential = [d for d in self.urldict if not self.urldict[d].all_visited]
-            if not potential:
-                self.done = True
-                return []
+        potential = self.get_unvisited_domains()
+        if not potential:
+            return []
         # variables init
         per_domain = max_urls // len(potential) or 1
         targets: List[Tuple[float, str]] = []
         # iterate potential domains
         for domain in potential:
-            url_tuples = self._load_urls(domain)
             # load urls
+            url_tuples = self._load_urls(domain)
             urlpaths: List[str] = []
             # get first non-seen urls
             for url in url_tuples:
@@ -339,17 +423,47 @@ class UrlStore:
             # store new info
             self._store_urls(domain, url_tuples, timestamp=total_diff)
         # sort by first tuple element (time in secs)
+        self._set_done()
         return sorted(targets, key=lambda x: x[0])  # type: ignore[arg-type]
 
+    # CRAWLING
+
+    def store_rules(self, website: str, rules: Optional[RobotFileParser]) -> None:
+        "Store crawling rules for a given website."
+        if self.compressed:
+            rules = zlib.compress(  # type: ignore[assignment]
+                pickle.dumps(rules, protocol=4)
+            )
+        self.urldict[website].rules = rules
+
     def get_rules(self, website: str) -> Optional[RobotFileParser]:
         "Return the stored crawling rules for the given website."
-        return self.urldict[website].rules
+        if website in self.urldict:
+            if self.compressed:
+                return pickle.loads(zlib.decompress(self.urldict[website].rules))  # type: ignore
+            return self.urldict[website].rules
+        return None
+
+    def get_crawl_delay(self, website: str, default: float = 5) -> float:
+        "Return the delay as extracted from robots.txt, or a given default."
+        delay = None
+        rules = self.get_rules(website)
+        try:
+            delay = rules.crawl_delay("*")  # type: ignore[union-attr]
+        except AttributeError:  # no rules or no crawl delay
+            pass
+        # backup
+        return delay or default  # type: ignore[return-value]
 
     # GENERAL INFO
 
+    def get_all_counts(self) -> List[int]:
+        "Return all download counts for the hosts in store."
+        return [self.urldict[d].count for d in self.urldict]
+
     def total_url_number(self) -> int:
         "Find number of all URLs in store."
-        return sum(len(self.urldict[d].tuples) for d in self.urldict)
+        return sum(self.urldict[d].total for d in self.urldict)
 
     def download_threshold_reached(self, threshold: float) -> bool:
         "Find out if the download limit (in seconds) has been reached for one of the websites in store."
@@ -358,18 +472,24 @@ class UrlStore:
     def dump_urls(self) -> List[str]:
         "Return a list of all known URLs."
         urls = []
-        for domain in self.get_known_domains():
+        for domain in self.urldict:
             urls.extend(self.find_known_urls(domain))
         return urls
 
+    def print_unvisited_urls(self) -> None:
+        "Print all unvisited URLs in store."
+        for domain in self.urldict:
+            print("\n".join(self.find_unvisited_urls(domain)), flush=True)
+
     def print_urls(self) -> None:
         "Print all URLs in store (URL + TAB + visited or not)."
-        for domain in self.get_known_domains():
+        for domain in self.urldict:
             print(
                 "\n".join(
                     [
                         domain + u.urlpath + "\t" + str(u.visited)
                         for u in self._load_urls(domain)
                     ]
-                )
+                ),
+                flush=True,
             )
