@@ -5,15 +5,26 @@ Functions related to URL manipulation and extraction of URL parts.
 import re
 
 from functools import lru_cache
+from html import unescape
 from typing import Any, List, Optional, Set, Tuple, Union
-from urllib.parse import urlparse, ParseResult
+from urllib.parse import urljoin, urlsplit, urlunsplit, SplitResult
 
 from tld import get_tld
 
 
-DOMAIN_REGEX = re.compile(r"(?:http|ftp)s?://(?:www[0-9]*\.)?([^/]+)")
+DOMAIN_REGEX = re.compile(
+    r"(?:(?:f|ht)tp)s?://"  # protocols
+    r"(?:[^/?#]{,63}\.)?"  # subdomain, www, etc.
+    r"([^/?#.]{4,63}\.[^/?#]{2,63}|"  # domain and extension
+    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|"  # IPv4
+    r"[0-9a-f:]{16,})"  # IPv6
+    r"(?:/|$)"  # slash or end of string
+)
 NO_EXTENSION_REGEX = re.compile(r"(^[^.]+)")
-CLEAN_DOMAIN_REGEX = re.compile(r"^www[0-9]*\.")
+STRIP_DOMAIN_REGEX = re.compile(r"^.+?:.*?@|(?<=\D):\d+")
+CLEAN_FLD_REGEX = re.compile(r"^www[0-9]*\.")
+INNER_SLASH_REGEX = re.compile(r"(.+/)+")
+FEED_WHITELIST_REGEX = re.compile(r"(?:feed(?:burner|proxy))", re.I)
 
 
 @lru_cache(maxsize=1024)
@@ -21,11 +32,13 @@ def get_tldinfo(
     url: str, fast: bool = False
 ) -> Union[Tuple[None, None], Tuple[str, str]]:
     """Cached function to extract top-level domain info"""
+    if not url or not isinstance(url, str):
+        return None, None
     if fast:
         # try with regexes
         domain_match = DOMAIN_REGEX.match(url)
         if domain_match:
-            full_domain = domain_match[1]
+            full_domain = STRIP_DOMAIN_REGEX.sub("", domain_match[1])
             clean_match = NO_EXTENSION_REGEX.match(full_domain)
             if clean_match:
                 return clean_match[0], full_domain
@@ -34,7 +47,7 @@ def get_tldinfo(
     if tldinfo is None:
         return None, None
     # this step is necessary to standardize output
-    return tldinfo.domain, CLEAN_DOMAIN_REGEX.sub("", tldinfo.fld)  # type: ignore[union-attr]
+    return tldinfo.domain, CLEAN_FLD_REGEX.sub("", tldinfo.fld)  # type: ignore[union-attr]
 
 
 def extract_domain(
@@ -55,11 +68,11 @@ def extract_domain(
     return full_domain
 
 
-def _parse(url: Any) -> ParseResult:
+def _parse(url: Any) -> SplitResult:
     "Parse a string or use urllib.parse object directly."
     if isinstance(url, str):
-        parsed_url = urlparse(url)
-    elif isinstance(url, ParseResult):
+        parsed_url = urlsplit(unescape(url))
+    elif isinstance(url, SplitResult):
         parsed_url = url
     else:
         raise TypeError("wrong input type:", type(url))
@@ -70,63 +83,59 @@ def get_base_url(url: Any) -> str:
     """Strip URL of some of its parts to get base URL.
     Accepts strings and urllib.parse ParseResult objects."""
     parsed_url = _parse(url)
-    return parsed_url._replace(path="", params="", query="", fragment="").geturl()
+    if parsed_url.scheme:
+        scheme = parsed_url.scheme + "://"
+    else:
+        scheme = ""
+    return scheme + parsed_url.netloc
 
 
 def get_host_and_path(url: Any) -> Tuple[str, str]:
     """Decompose URL in two parts: protocol + host/domain and path.
     Accepts strings and urllib.parse ParseResult objects."""
     parsed_url = _parse(url)
-    host = parsed_url._replace(path="", params="", query="", fragment="")
-    path = parsed_url._replace(scheme="", netloc="")
-    hostval, pathval = host.geturl(), path.geturl()
+    hostname = get_base_url(parsed_url)
+    pathval = urlunsplit(
+        ["", "", parsed_url.path, parsed_url.query, parsed_url.fragment]
+    )
     # correction for root/homepage
     if pathval == "":
         pathval = "/"
-    if not hostval or not pathval:
+    if not hostname or not pathval:
         raise ValueError(f"incomplete URL: {url}")
-    return hostval, pathval
+    return hostname, pathval
 
 
 def get_hostinfo(url: str) -> Tuple[Optional[str], str]:
-    "Extract domain and host info (protocol + host/domain) from a URL."
+    "Convenience function returning domain and host info (protocol + host/domain) from a URL."
     domainname = extract_domain(url, fast=True)
-    parsed_url = urlparse(url)
-    host = parsed_url._replace(path="", params="", query="", fragment="")
-    return domainname, host.geturl()
+    base_url = get_base_url(url)
+    return domainname, base_url
 
 
 def fix_relative_urls(baseurl: str, url: str) -> str:
     "Prepend protocol and host information to relative links."
-    if url.startswith("//"):
-        return "https:" + url if baseurl.startswith("https") else "http:" + url
-    if url.startswith("/"):
-        # imperfect path handling
-        return baseurl + url
-    if url.startswith("."):
-        # don't try to correct these URLs
-        return baseurl + "/" + re.sub(r"(.+/)+", "", url)
-    if not url.startswith("http") and not url.startswith("{"):
-        return baseurl + "/" + url
-    # todo: handle here
-    # if url.startswith('{'):
-    return url
+    if url.startswith("{"):
+        return url
+    base_netloc = urlsplit(baseurl).netloc
+    split_url = urlsplit(url)
+    if split_url.netloc not in (base_netloc, ""):
+        if split_url.scheme:
+            return url
+        return urlunsplit(split_url._replace(scheme="http"))
+    return urljoin(baseurl, url)
 
 
-def filter_urls(linklist: List[str], urlfilter: Optional[str]) -> List[str]:
+def filter_urls(link_list: List[str], urlfilter: Optional[str]) -> List[str]:
     "Return a list of links corresponding to the given substring pattern."
     if urlfilter is None:
-        return sorted(set(linklist))
+        return sorted(set(link_list))
     # filter links
-    newlist = [l for l in linklist if urlfilter in l]
+    filtered_list = [l for l in link_list if urlfilter in l]
     # feedburner option: filter and wildcards for feeds
-    if not newlist:
-        newlist = [
-            l
-            for l in linklist
-            if urlfilter in l or "feedburner" in l or "feedproxy" in l
-        ]
-    return sorted(set(newlist))
+    if not filtered_list:
+        filtered_list = [l for l in link_list if FEED_WHITELIST_REGEX.search(l)]
+    return sorted(set(filtered_list))
 
 
 def is_external(url: str, reference: str, ignore_suffix: bool = True) -> bool:
@@ -142,20 +151,25 @@ def is_external(url: str, reference: str, ignore_suffix: bool = True) -> bool:
 
 def is_known_link(link: str, known_links: Set[str]) -> bool:
     "Compare the link and its possible variants to the existing URL base."
-    # easy check
+    # check exact link
     if link in known_links:
         return True
-    # trailing slash
-    test1 = link.rstrip("/")
-    test2 = test1 + "/"
-    if test1 in known_links or test2 in known_links:
+
+    # check link and variants with trailing slashes
+    test_links = [link.rstrip("/"), link.rstrip("/") + "/"]
+    if any(test_link in known_links for test_link in test_links):
         return True
-    # http/https + trailing slash
+
+    # check link and variants with modified protocol
     if link.startswith("http"):
         if link.startswith("https"):
             testlink = link[:4] + link[5:]
         else:
             testlink = "".join([link[:4], "s", link[4:]])
-        test1, test2 = testlink.rstrip("/"), testlink.rstrip("/") + "/"
-        return testlink in known_links or test1 in known_links or test2 in known_links
+        if any(
+            test in known_links
+            for test in [testlink, testlink.rstrip("/"), testlink.rstrip("/") + "/"]
+        ):
+            return True
+
     return False
