@@ -14,7 +14,7 @@ import tempfile
 
 from contextlib import redirect_stdout
 from unittest.mock import patch
-from urllib.parse import ParseResult
+from urllib.parse import SplitResult, urlsplit
 
 import pytest
 
@@ -27,6 +27,7 @@ from courlan import (
     is_external,
     sample_urls,
     validate_url,
+    is_valid_url,
     extract_links,
     extract_domain,
     filter_urls,
@@ -38,8 +39,10 @@ from courlan import (
     is_not_crawlable,
     lang_filter,
 )
-from courlan.filters import extension_filter, path_filter, type_filter
-from courlan.urlutils import _parse, is_known_link
+from courlan.core import filter_links
+from courlan.filters import domain_filter, extension_filter, path_filter, type_filter
+from courlan.meta import clear_caches
+from courlan.urlutils import _parse, get_tldinfo, is_known_link
 
 
 logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
@@ -97,12 +100,60 @@ def test_fix_relative():
         fix_relative_urls("https://example.org", "../../test.html")
         == "https://example.org/test.html"
     )
+    # sub-directories
+    assert (
+        fix_relative_urls("https://www.example.org/dir/subdir/file.html", "/absolute")
+        == "https://www.example.org/absolute"
+    )
+    assert (
+        fix_relative_urls("https://www.example.org/dir/subdir/file.html", "relative")
+        == "https://www.example.org/dir/subdir/relative"
+    )
+    assert (
+        fix_relative_urls("https://www.example.org/dir/subdir/", "relative")
+        == "https://www.example.org/dir/subdir/relative"
+    )
+    assert (
+        fix_relative_urls("https://www.example.org/dir/subdir", "relative")
+        == "https://www.example.org/dir/relative"
+    )
+    # non-relative URLs
+    assert (
+        fix_relative_urls("https://example.org", "https://www.eff.org")
+        == "https://www.eff.org"
+    )
+    assert (
+        fix_relative_urls("https://example.org", "//www.eff.org")
+        == "http://www.eff.org"
+    )
+    # looks like an absolute URL but is actually a valid relative URL
+    assert (
+        fix_relative_urls("https://example.org", "www.eff.org")
+        == "https://example.org/www.eff.org"
+    )
+    # misc
+    assert (
+        fix_relative_urls("https://www.example.org/dir/subdir/file.html", "./this:that")
+        == "https://www.example.org/dir/subdir/this:that"
+    )
+    assert (
+        fix_relative_urls(
+            "https://www.example.org/test.html?q=test#frag", "foo.html?q=bar#baz"
+        )
+        == "https://www.example.org/foo.html?q=bar#baz"
+    )
+    assert fix_relative_urls("https://www.example.org", "{privacy}") == "{privacy}"
 
 
 def test_scrub():
     # clean: scrub + normalize
     assert clean_url(5) is None
-    assert clean_url("ø\xaa") == "øª"
+    assert clean_url("ø\xaa") == "%C3%B8%C2%AA"
+    assert clean_url("https://example.org/?p=100") == "https://example.org/?p=100"
+    assert (
+        clean_url("https://example.org:443/file.html?p=100&abc=1#frag")
+        == "https://example.org/file.html?abc=1&p=100#frag"
+    )
     # scrub
     assert scrub_url("  https://www.dwds.de") == "https://www.dwds.de"
     assert scrub_url("<![CDATA[https://www.dwds.de]]>") == "https://www.dwds.de"
@@ -139,11 +190,13 @@ def test_scrub():
     assert scrub_url("\x19https://www.test.com/\x06") == "https://www.test.com"
     # markup
     assert scrub_url("https://www.test.com/</a>") == "https://www.test.com"
+    assert scrub_url("https://www.test.com/1</div>") == "https://www.test.com/1"
+    assert scrub_url("https://www.test.com/{user_name}") == "https://www.test.com"
     # garbled URLs e.g. due to quotes
     assert (
         scrub_url('https://www.test.com/"' + "<p></p>" * 100) == "https://www.test.com"
     )
-    assert scrub_url('https://www.test.com/"' * 50) != "https://www.test.com"
+    assert scrub_url('https://www.test.com/"' * 50) == "https://www.test.com"
     # simply too long, left untouched
     my_url = "https://www.test.com/" + "abcdefg" * 100
     assert scrub_url(my_url) == my_url
@@ -181,8 +234,12 @@ def test_extension_filter():
 
 
 def test_spam_filter():
-    assert type_filter("http://www.example.org/cams/test.html", strict=False) is True
-    assert type_filter("http://www.example.org/cams/test.html", strict=True) is False
+    assert (
+        type_filter("http://www.example.org/livecams/test.html", strict=False) is True
+    )
+    assert (
+        type_filter("http://www.example.org/livecams/test.html", strict=True) is False
+    )
     assert type_filter("http://www.example.org/test.html") is True
 
 
@@ -190,31 +247,51 @@ def test_type_filter():
     assert type_filter("http://www.example.org/feed") is False
     # straight category
     assert type_filter("http://www.example.org/category/123") is False
+    assert type_filter("http://www.example.org/product-category/123") is False
     # post simply filed under a category
     assert type_filter("http://www.example.org/category/tropes/time-travel") is True
     assert (
         type_filter("http://www.example.org/test.xml?param=test", strict=True) is False
     )
     assert type_filter("http://www.example.org/test.asp") is True
-    assert type_filter("http://ads.example.org/") is False
     # -video- vs. /video/
-    assert type_filter("http://my-videos.com/") is True
-    assert type_filter("http://my-videos.com/", strict=True) is False
-    assert type_filter("http://example.com/video/1", strict=True) is False
-    assert type_filter("http://example.com/new-video-release") is True
-    assert type_filter("http://example.com/new-video-release", strict=True) is False
+    assert type_filter("http://my-livechat.com/") is True
+    assert type_filter("http://my-livechat.com/", strict=True) is False
+    assert type_filter("http://example.com/livechat/1", strict=True) is False
+    assert type_filter("http://example.com/new-sexcam") is True
+    assert type_filter("http://example.com/new-sexcam", strict=True) is False
     # tags
     assert type_filter("https://de.thecitizen.de/tag/anonymity/") is False
     assert type_filter("https://de.thecitizen.de/tags/anonymity/") is False
     # author
     assert type_filter("http://www.example.org/author/abcde") is False
     assert type_filter("http://www.example.org/autor/abcde/") is False
+    # archives
+    assert type_filter("http://www.example.org/2011/11/") is False
+    assert type_filter("http://www.example.org/2011/") is False
+    assert type_filter("http://www.example.org/2011_archive.html") is False
+    assert type_filter("http://www.example.org/2020/02/06/1859/") is True
     # misc
     assert (
         type_filter("http://www.bmbwk.gv.at/forschung/fps/gsk/befragung.xml?style=text")
         is True
     )
-    # assert type_filter('http://www.aec.at/de/archives/prix_archive/prix_projekt.asp?iProjectID=11118') is True
+    assert (
+        type_filter(
+            "http://www.aec.at/de/archives/prix_archive/prix_projekt.asp?iProjectID=11118"
+        )
+        is False
+    )
+    # nav
+    assert type_filter("http://www.example.org/tag/abcde/", with_nav=False) is False
+    assert type_filter("http://www.example.org/tag/abcde/", with_nav=True) is True
+    assert type_filter("http://www.example.org/page/10/", with_nav=False) is False
+    assert type_filter("http://www.example.org/page/10/", with_nav=True) is True
+    # img
+    assert type_filter("http://www.example.org/logo_800_web-jpg/", strict=True) is False
+    assert (
+        type_filter("http://www.example.org/img_2020-03-03_25/", strict=True) is False
+    )
 
 
 def test_path_filter():
@@ -224,10 +301,18 @@ def test_path_filter():
         )
         is not None
     )
+    assert path_filter("/index.php", "") is False
     assert check_url("http://www.case-modder.de/index.php", strict=True) is None
+    assert path_filter("/default/", "") is False
+    assert check_url("http://www.case-modder.de/default/", strict=True) is None
+    assert path_filter("/contact/", "") is False
+    assert path_filter("/Datenschutzerklaerung", "") is False
+    # assert path_filter("/", "") is False
 
 
 def test_lang_filter():
+    assert lang_filter("http://test.com/az/", "de") is False
+    assert lang_filter("http://test.com/de/", "de") is True
     assert (
         lang_filter(
             "https://www.20min.ch/fr/story/des-millions-pour-produire-de-l-energie-renouvelable-467974085377",
@@ -377,6 +462,10 @@ def test_validate():
     assert validate_url("ftps://www.test.org/test")[0] is False
     assert validate_url("http://t.g/test")[0] is False
     assert validate_url("http://test.org/test")[0] is True
+    # assert validate_url("http://sub.-mkyong.com/test")[0] is False
+
+    assert not is_valid_url("http://www.test[.org/test")
+    assert is_valid_url("http://test.org/test")
 
 
 def test_normalization():
@@ -390,35 +479,80 @@ def test_normalization():
         == "http://test.net/foo.html#bar"
     )
     assert (
-        normalize_url("http://test.net/foo.html#:~:text=night-,vision", strict=True)
-        == "http://test.net/foo.html"
+        normalize_url("http://test.net/foo.html#:~:text=night-,vision")
+        == "http://test.net/foo.html#:~:text=night-,vision"
     )
     assert (
         normalize_url("http://www.example.org:80/test.html")
         == "http://www.example.org/test.html"
     )
+    assert (
+        normalize_url("http://www.example.org:80?p=123")
+        == "http://www.example.org/?p=123"
+    )
     assert (
         normalize_url("https://hanxiao.io//404.html") == "https://hanxiao.io/404.html"
     )
 
+    # punycode
+    assert normalize_url("http://xn--Mnchen-3ya.de") == "http://münchen.de"
+    assert normalize_url("http://Mnchen-3ya.de") == "http://mnchen-3ya.de"
+    assert normalize_url("http://xn--München.de") == "http://xn--münchen.de"
+
+    # account for particular characters
+    assert (
+        normalize_url(
+            "https://www.deutschlandfunknova.de/beitrag/nord--und-s%C3%BCdgaza-israels-armee-verk%C3%BCndet-teilung-des-gazastreifens"
+        )
+        == "https://www.deutschlandfunknova.de/beitrag/nord--und-s%C3%BCdgaza-israels-armee-verk%C3%BCndet-teilung-des-gazastreifens"
+    )
+    assert (
+        normalize_url("https://taz.de/Zukunft-des-49-Euro-Tickets/!5968518/")
+        == "https://taz.de/Zukunft-des-49-Euro-Tickets/!5968518/"
+    )
+
+    # trackers
+    assert normalize_url("http://test.org/?s_cid=123&clickid=1") == "http://test.org/"
+    assert normalize_url("http://test.org/?aftr_source=0") == "http://test.org/"
+    assert normalize_url("http://test.org/?fb_ref=0") == "http://test.org/"
+    assert normalize_url("http://test.org/?this_affiliate=0") == "http://test.org/"
+    assert (
+        normalize_url("http://test.org/?utm_source=rss&utm_medium=rss")
+        == "http://test.org/"
+    )
+    assert (
+        normalize_url("http://test.org/?utm_source=rss&#038;utm_medium=rss")
+        == "http://test.org/"
+    )
+    assert normalize_url("http://test.org/#partnerid=123") == "http://test.org/"
+    assert (
+        normalize_url(
+            "http://test.org/#mtm_campaign=documentation&mtm_keyword=demo&catpage=3"
+        )
+        == "http://test.org/#catpage=3"
+    )
+    assert normalize_url("http://test.org/#page2") == "http://test.org/#page2"
+
 
 def test_qelems():
     assert (
         normalize_url("http://test.net/foo.html?utm_source=twitter")
-        == "http://test.net/foo.html?utm_source=twitter"
-    )
-    assert (
-        normalize_url("http://test.net/foo.html?utm_source=twitter", strict=True)
         == "http://test.net/foo.html"
     )
     assert (
-        normalize_url("http://test.net/foo.html?utm_source=twitter&post=abc&page=2")
-        == "http://test.net/foo.html?page=2&post=abc&utm_source=twitter"
+        normalize_url("http://test.net/foo.html?testid=1")
+        == "http://test.net/foo.html?testid=1"
     )
     assert (
-        normalize_url(
-            "http://test.net/foo.html?utm_source=twitter&post=abc&page=2", strict=True
-        )
+        normalize_url("http://test.net/foo.html?testid=1", strict=True)
+        == "http://test.net/foo.html"
+    )
+    assert (
+        normalize_url("http://test.net/foo.html?testid=1&post=abc&page=2")
+        == "http://test.net/foo.html?page=2&post=abc&testid=1"
+    )
+    assert (
+        normalize_url("http://test.net/foo.html?testid=1&post=abc&page=2", strict=True)
         == "http://test.net/foo.html?page=2&post=abc"
     )
     assert (
@@ -450,20 +584,14 @@ def test_urlcheck():
     assert check_url("http://example.com/test.js") is None
     assert check_url("http://twitter.com/", strict=True) is None
     assert check_url("http://twitter.com/", strict=False) is not None
-    assert check_url("https://www.httpbin.org/status/200", with_redirects=True) == (
-        "https://www.httpbin.org/status/200",
-        "httpbin.org",
-    )
-    # assert check_url('https://www.httpbin.org/status/302', with_redirects=True) == ('https://www.httpbin.org/status/302', 'httpbin.org')
-    assert check_url("https://www.httpbin.org/status/404", with_redirects=True) is None
-    assert check_url("https://www.ht.or", with_redirects=True) is None
+
     # recheck type and spam filters
-    assert check_url("http://example.org/code/oembed/") is None
-    assert check_url("http://cams.com/", strict=False) == (
-        "http://cams.com",
-        "cams.com",
+    assert check_url("http://example.org/wp-json/oembed/") is None
+    assert check_url("http://livecams.com/", strict=False) == (
+        "http://livecams.com",
+        "livecams.com",
     )
-    assert check_url("http://cams.com/", strict=True) is None
+    assert check_url("http://livecams.com/", strict=True) is None
     assert check_url("https://denkiterm.wordpress.com/impressum/", strict=True) is None
     assert (
         check_url(
@@ -531,18 +659,18 @@ def test_urlcheck():
             language="de",
             strict=True,
         )
-        is not None
+        is None
     )
     assert (
         check_url(
-            "https://de.nachrichten.yahoo.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html",
+            "https://de.nachrichten.other.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html",
             language="en",
         )
         is not None
     )
     assert (
         check_url(
-            "https://de.nachrichten.yahoo.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html",
+            "https://de.nachrichten.other.com/bundesliga-schiri-boss-fr%C3%B6hlich-f%C3%BCr-175850830.html",
             language="en",
             strict=True,
         )
@@ -550,17 +678,96 @@ def test_urlcheck():
     )
     # assert check_url('http://www.immobilienscout24.de/de/ueberuns/presseservice/pressestimmen/2_halbjahr_2000.jsp;jsessionid=287EC625A45BD5A243352DD8C86D25CC.worker2', language='de', strict=True) is not None
 
+    # domain name
+    assert check_url("http://-100x100.webp") is None
+    assert check_url("http://0.gravata.html") is None
+    assert check_url("http://https:") is None
+    assert check_url("http://127.0.0.1") is not None
+    assert check_url("http://111.111.111.111") is not None
+    assert check_url("http://0127.0.0.1") is None
+    # assert check_url("http://::1") is not None
+    assert check_url("http://2001:0db8:85a3:0000:0000:8a2e:0370:7334") is not None
+    assert check_url("http://[2001:0db8:85a3:0000:0000:8a2e:0370:7334]") is None
+    assert check_url("http://1:2:3:4:5:6:7:8:9") is None
+
+
+def test_domain_filter():
+    "Test filters related to domain and hostnames."
+    assert domain_filter("") is False
+    assert domain_filter("too-long" + "g" * 60 + ".org") is False
+    assert domain_filter("long" + "g" * 50 + ".org") is True
+    assert domain_filter("example.-com") is False
+    assert domain_filter("example.") is False
+    assert domain_filter("-example.com") is False
+    assert domain_filter("_example.com") is False
+    assert domain_filter("example.com:") is False
+    assert domain_filter("a......b.com") is False
+    assert domain_filter("*.example.com") is False
+    assert domain_filter("exa-mple.co.uk") is True
+    assert domain_filter("kräuter.de") is True
+    assert domain_filter("xn--h1aagokeh.xn--p1ai") is True
+    assert domain_filter("`$smarty.server.server_name`") is False
+    assert domain_filter("$`)}if(a.tryconvertencoding)trycatch(e)const") is False
+    assert domain_filter("00x200.jpg,") is False
+    assert domain_filter("-100x100.webp") is False
+    assert domain_filter("0.gravata.html") is False
+    assert domain_filter("https:") is False
+
+    assert domain_filter("127.0.0.1") is True
+    assert domain_filter("900.200.100.75") is False
+    assert domain_filter("111.111.111") is False
+    assert domain_filter("0127.0.0.1") is False
+
+    assert domain_filter("example.jpg") is False
+    assert domain_filter("example.html") is False
+    assert domain_filter("0.gravatar.com") is False
+    assert domain_filter("12345.org") is False
+    # assert domain_filter("test.invalidtld") is False
+
+
+def test_urlcheck_redirects():
+    "Test redirection checks."
+    assert check_url("https://www.httpbun.org/status/200", with_redirects=True) == (
+        "https://httpbun.org",
+        "httpbun.org",
+    )
+    assert check_url("https://www.httpbin.org/status/404", with_redirects=True) is None
+    assert check_url("https://www.ht.or", with_redirects=True) is None
+
 
 def test_urlutils():
     """Test URL manipulation tools"""
     # domain extraction
+    assert extract_domain("") is None
+    assert extract_domain(5) is None
     assert extract_domain("h") is None
     assert extract_domain("https://httpbin.org/") == "httpbin.org"
+    assert extract_domain("https://www.httpbin.org/", fast=True) == "httpbin.org"
+    assert extract_domain("http://www.mkyong.com.au", fast=True) == "mkyong.com.au"
+    assert extract_domain("http://mkyong.t.t.co", fast=True) == "mkyong.t.t.co"
+    assert extract_domain("ftp://www4.httpbin.org", fast=True) == "httpbin.org"
+    assert extract_domain("http://w3.example.com", fast=True) == "example.com"
+    assert extract_domain("https://de.nachrichten.yahoo.com/", fast=True) == "yahoo.com"
+    assert (
+        extract_domain("http://xn--h1aagokeh.xn--p1ai:8888", fast=True)
+        == "xn--h1aagokeh.xn--p1ai"
+    )
+    assert extract_domain("http://user:pass@domain.test:81", fast=True) == "domain.test"
+    assert extract_domain("http://111.2.33.44/test", fast=True) == "111.2.33.44"
+    assert (
+        extract_domain("http://2001:db8::ff00:42:8329/test", fast=True)
+        == "2001:db8::ff00:42:8329"
+    )
+    assert (
+        extract_domain("https://test.xn--0zwm56d.com/", fast=True) == "xn--0zwm56d.com"
+    )
+    assert extract_domain("http://example.com?query=one", fast=True) == "example.com"
+    assert extract_domain("http://example.com#fragment", fast=True) == "example.com"
     # url parsing
     result = _parse("https://httpbin.org/")
-    assert isinstance(result, ParseResult)
+    assert isinstance(result, SplitResult)
     newresult = _parse(result)
-    assert isinstance(result, ParseResult)
+    assert isinstance(result, SplitResult)
     with pytest.raises(TypeError):
         result = _parse(1.23)
 
@@ -600,6 +807,7 @@ def test_urlutils():
 
 def test_external():
     """test domain comparison"""
+    assert is_external("", "https://www.microsoft.com/") is True
     assert is_external("https://github.com/", "https://www.microsoft.com/") is True
     assert (
         is_external(
@@ -636,6 +844,9 @@ def test_extraction():
     # link known under another form
     pagecontent = '<html><a href="https://test.org/example"/><a href="https://test.org/example/&"/></html>'
     assert len(extract_links(pagecontent, "https://test.org", False)) == 1
+    # nofollow
+    pagecontent = '<html><a href="https://test.com/example" rel="nofollow ugc"/></html>'
+    assert len(extract_links(pagecontent, "https://test.com/", False)) == 0
     # language
     pagecontent = '<html><a href="https://test.com/example" hreflang="de-DE"/></html>'
     assert len(extract_links(pagecontent, "https://test.com/", False)) == 1
@@ -662,40 +873,139 @@ def test_extraction():
     )
     # language + content
     pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/example2"/><a href="https://test.com/example2 ADDITIONAL"/></html>'
-    links = extract_links(pagecontent, "https://test.com/", False)
+    links = extract_links(pagecontent, "https://test.com/", external_bool=False)
     assert sorted(links) == ["https://test.com/example", "https://test.com/example2"]
     assert (
-        len(extract_links(pagecontent, "https://test.com/", False, language="de")) == 2
+        len(
+            extract_links(
+                pagecontent, "https://test.com/", external_bool=False, language="de"
+            )
+        )
+        == 2
     )
     pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="https://test.com/page/2"/></html>'
     assert (
-        len(extract_links(pagecontent, "https://test.com/", False, with_nav=False)) == 1
+        len(
+            extract_links(
+                pagecontent, "https://test.com/", external_bool=False, with_nav=False
+            )
+        )
+        == 1
     )
     assert (
-        len(extract_links(pagecontent, "https://test.com/", False, with_nav=True)) == 2
+        len(
+            extract_links(
+                pagecontent, "https://test.com/", external_bool=False, with_nav=True
+            )
+        )
+        == 2
     )
+    # navigation
     pagecontent = "<html><head><title>Links</title></head><body><a href='/links/2/0'>0</a> <a href='/links/2/1'>1</a> </body></html>"
-    links = extract_links(pagecontent, "https://httpbin.org", False, with_nav=True)
+    links = extract_links(
+        pagecontent, "https://httpbin.org", external_bool=False, with_nav=True
+    )
     assert sorted(links) == [
         "https://httpbin.org/links/2/0",
         "https://httpbin.org/links/2/1",
     ]
+    links = extract_links(
+        pagecontent, base_url="https://httpbin.org", external_bool=False, with_nav=True
+    )
+    assert sorted(links) == [
+        "https://httpbin.org/links/2/0",
+        "https://httpbin.org/links/2/1",
+    ]
+    pagecontent = "<html><head><title>Links</title></head><body><a href='links/2/0'>0</a> <a href='links/2/1'>1</a> </body></html>"
+    links = extract_links(
+        pagecontent,
+        url="https://httpbin.org/page1/",
+        external_bool=False,
+        with_nav=True,
+    )
+    assert sorted(links) == [
+        "https://httpbin.org/page1/links/2/0",
+        "https://httpbin.org/page1/links/2/1",
+    ]
+    pagecontent = "<html><head><title>Pages</title></head><body><a href='/page/10'>10</a> <a href='/page/?=11'>11</a></body></html>"
+    assert (
+        extract_links(
+            pagecontent,
+            "https://example.org",
+            external_bool=False,
+            strict=False,
+            with_nav=False,
+        )
+        == set()
+    )
+    links = extract_links(
+        pagecontent,
+        "https://example.org",
+        external_bool=False,
+        strict=True,
+        with_nav=True,
+    )
+    assert sorted(links) == [
+        "https://example.org/page/",  # parameter stripped by strict filtering
+        "https://example.org/page/10",
+    ]
+    links = extract_links(
+        pagecontent,
+        "https://example.org",
+        external_bool=False,
+        strict=False,
+        with_nav=True,
+    )
+    assert sorted(links) == [
+        "https://example.org/page/10",
+        "https://example.org/page/?=11",
+    ]
     # links undeveloped by CMS
     pagecontent = (
         '<html><a href="{privacy}" target="_privacy">{privacy-link}</a></html>'
     )
-    assert len(extract_links(pagecontent, "https://test.com/", False)) == 0
-    assert len(extract_links(pagecontent, "https://test.com/", True)) == 0
+    assert (
+        len(extract_links(pagecontent, "https://test.com/", external_bool=False)) == 0
+    )
+    assert len(extract_links(pagecontent, "https://test.com/", external_bool=True)) == 0
     # links without quotes
-    pagecontent = "<html><a href=/contact>Link</a></html>"
-    assert extract_links(pagecontent, "https://test.com/", False) == {
-        "https://test.com/contact"
+    pagecontent = "<html><a href=/link>Link</a></html>"
+    assert extract_links(pagecontent, "https://test.com/", external_bool=False) == {
+        "https://test.com/link"
     }
-    assert extract_links(pagecontent, "https://test.com/", True) == set()
-    pagecontent = "<html><a href=/contact attribute=value>Link</a></html>"
-    assert extract_links(pagecontent, "https://test.com/", False) == {
-        "https://test.com/contact"
+    assert extract_links(pagecontent, "https://test.com/", external_bool=True) == set()
+    pagecontent = "<html><a href=/link attribute=value>Link</a></html>"
+    assert extract_links(pagecontent, "https://test.com/", external_bool=False) == {
+        "https://test.com/link"
     }
+    # external links with extension (here ".com")
+    pagecontent = '<html><body><a href="https://knoema.com/o/data-engineer-india"/><a href="https://knoema.recruitee.com/"/></body></html>'
+    assert extract_links(pagecontent, "https://knoema.com/", external_bool=False) == {
+        "https://knoema.com/o/data-engineer-india"
+    }
+    assert extract_links(pagecontent, "https://knoema.com/", external_bool=True) == {
+        "https://knoema.recruitee.com"
+    }
+    # return all links without filters
+    pagecontent = '<html><a hreflang="de-DE" href="https://test.com/example"/><a href="/page/2"/><a href="https://example.com/gallery/"/></html>'
+    result = extract_links(
+        pagecontent, "https://test.com", external_bool=True, no_filter=True
+    )
+    assert sorted(result) == [
+        "https://example.com/gallery/",
+        "https://test.com/example",
+        "https://test.com/page/2",
+    ]
+    # link filtering
+    base_url = "https://example.org"
+    htmlstring = '<html><body><a href="https://example.org/page1"/><a href="https://example.org/page1/"/><a href="https://test.org/page1"/></body></html>'
+    links, links_priority = filter_links(htmlstring, base_url)
+    assert len(links) == 1 and not links_priority
+    # link filtering with relative URLs
+    url = "https://example.org/page1.html"
+    htmlstring = '<html><body><a href="/subpage1"/><a href="/subpage1/"/><a href="https://test.org/page1"/></body></html>'
+    links, links_priority = filter_links(htmlstring, url=url)
+    assert len(links) == 1 and not links_priority
 
 
 def test_cli():
@@ -711,6 +1021,8 @@ def test_cli():
         "-v",
         "--language",
         "en",
+        "--parallel",
+        "2",
     ]
     with patch.object(sys, "argv", testargs):
         args = cli.parse_args(testargs)
@@ -719,8 +1031,15 @@ def test_cli():
     assert args.outputfile == "output.txt"
     assert args.verbose is True
     assert args.language == "en"
+    assert args.parallel == 2
     assert os.system("courlan --help") == 0  # exit status
 
+    # _cli_check_urls
+    assert cli._cli_check_urls(["123", "https://example.org"]) == [
+        (False, "123"),
+        (True, "https://example.org"),
+    ]
+
     # testfile
     inputfile = os.path.join(RESOURCES_DIR, "input.txt")
     os_handle, temp_outputfile = tempfile.mkstemp(suffix=".txt", text=True)
@@ -734,7 +1053,7 @@ def test_cli():
     # test for Windows and the rest
     assert (
         subprocess.run(
-            [courlan_bin, "-i", inputfile, "-o", temp_outputfile], env=env
+            [courlan_bin, "-i", inputfile, "-o", temp_outputfile, "-p", "1"], env=env
         ).returncode
         == 0
     )
@@ -753,6 +1072,8 @@ def test_cli():
             "--language",
             "en",
             "--strict",
+            "-p",
+            "1",
         ]
         f = io.StringIO()
         with patch.object(sys, "argv", testargs):
@@ -760,12 +1081,17 @@ def test_cli():
         with redirect_stdout(f):
             cli.process_args(args)
         assert len(f.getvalue()) == 0
-        testargs = ["", "-i", inputfile, "-o", "/tmp/tralala.txt", "--sample"]
+
+        testargs = ["", "-i", inputfile, "-o", "/tmp/tralala.txt", "--sample", "10"]
         with patch.object(sys, "argv", testargs):
             args = cli.parse_args(testargs)
         with redirect_stdout(f):
             cli.process_args(args)
         assert len(f.getvalue()) == 0
+        args.verbose = True
+        with redirect_stdout(f):
+            cli.process_args(args)
+        assert len(f.getvalue()) == 0
     # delete temporary output file
     try:
         os.remove(temp_outputfile)
@@ -776,6 +1102,7 @@ def test_cli():
 def test_sample():
     """test URL sampling"""
     assert not list(sample_urls(["http://test.org/test1", "http://test.org/test2"], 0))
+    assert not list(sample_urls(["http://test.org/", "http://test.org"], 10))
 
     # assert len(sample_urls(['http://test.org/test1', 'http://test.org/test2'], 1)) == 1
     mylist = [
@@ -808,6 +1135,11 @@ def test_examples():
         "https://github.com/adbar/courlan",
         "github.com",
     )
+    assert check_url("http://666.0.0.1/") is None
+    assert check_url("http://test.net/foo.html?utm_source=twitter#gclid=123") == (
+        "http://test.net/foo.html",
+        "test.net",
+    )
     assert check_url(
         "https://httpbin.org/redirect-to?url=http%3A%2F%2Fexample.org", strict=True
     ) == ("https://httpbin.org/redirect-to", "httpbin.org")
@@ -821,3 +1153,20 @@ def test_examples():
         )
         == "http://test.net/foo.html?page=2&post=abc"
     )
+
+
+def test_meta():
+    "Test package meta functions."
+    url = "https://example.net/123/abc"
+    _ = get_tldinfo(url)
+    _ = _parse(url)
+    assert get_tldinfo.cache_info().currsize > 0
+    try:
+        urlsplit_lrucache = True
+        assert urlsplit.cache_info().currsize > 0
+    except AttributeError:  # newer Python versions only
+        urlsplit_lrucache = False
+    clear_caches()
+    assert get_tldinfo.cache_info().currsize == 0
+    if urlsplit_lrucache:
+        assert urlsplit.cache_info().currsize == 0
